export type PDFSidebarOptions = {
    /**
     * - The DOM elements.
     */
    elements: PDFSidebarElements;
    /**
     * - The document viewer.
     */
    pdfViewer: any;
    /**
     * - The thumbnail viewer.
     */
    pdfThumbnailViewer: any;
    /**
     * - The application event bus.
     */
    eventBus: any;
    /**
     * - The localization service.
     */
    l10n: any;
    /**
     * - Disable the notification for
     * documents containing outline/attachments. The default value is `false`.
     */
    disableNotification?: boolean | undefined;
};
export type PDFSidebarElements = {
    /**
     * - The outer container
     * (encasing both the viewer and sidebar elements).
     */
    outerContainer: HTMLDivElement;
    /**
     * - The viewer container
     * (in which the viewer element is placed).
     */
    viewerContainer: HTMLDivElement;
    /**
     * - The button used for
     * opening/closing the sidebar.
     */
    toggleButton: HTMLButtonElement;
    /**
     * - The button used to show
     * the thumbnail view.
     */
    thumbnailButton: HTMLButtonElement;
    /**
     * - The button used to show
     * the outline view.
     */
    outlineButton: HTMLButtonElement;
    /**
     * - The button used to show
     * the attachments view.
     */
    attachmentsButton: HTMLButtonElement;
    /**
     * - The button used to show
     * the layers view.
     */
    layersButton: HTMLButtonElement;
    /**
     * - The container in which
     * the thumbnails are placed.
     */
    thumbnailView: HTMLDivElement;
    /**
     * - The container in which
     * the outline is placed.
     */
    outlineView: HTMLDivElement;
    /**
     * - The container in which
     * the attachments are placed.
     */
    attachmentsView: HTMLDivElement;
    /**
     * - The container in which
     * the layers are placed.
     */
    layersView: HTMLDivElement;
};
export namespace SidebarView {
    const UNKNOWN: number;
    const NONE: number;
    const THUMBS: number;
    const OUTLINE: number;
    const ATTACHMENTS: number;
    const LAYERS: number;
}
/**
 * @typedef {Object} PDFSidebarOptions
 * @property {PDFSidebarElements} elements - The DOM elements.
 * @property {PDFViewer} pdfViewer - The document viewer.
 * @property {PDFThumbnailViewer} pdfThumbnailViewer - The thumbnail viewer.
 * @property {EventBus} eventBus - The application event bus.
 * @property {IL10n} l10n - The localization service.
 * @property {boolean} [disableNotification] - Disable the notification for
 *   documents containing outline/attachments. The default value is `false`.
 */
/**
 * @typedef {Object} PDFSidebarElements
 * @property {HTMLDivElement} outerContainer - The outer container
 *   (encasing both the viewer and sidebar elements).
 * @property {HTMLDivElement} viewerContainer - The viewer container
 *   (in which the viewer element is placed).
 * @property {HTMLButtonElement} toggleButton - The button used for
 *   opening/closing the sidebar.
 * @property {HTMLButtonElement} thumbnailButton - The button used to show
 *   the thumbnail view.
 * @property {HTMLButtonElement} outlineButton - The button used to show
 *   the outline view.
 * @property {HTMLButtonElement} attachmentsButton - The button used to show
 *   the attachments view.
 * @property {HTMLButtonElement} layersButton - The button used to show
 *   the layers view.
 * @property {HTMLDivElement} thumbnailView - The container in which
 *   the thumbnails are placed.
 * @property {HTMLDivElement} outlineView - The container in which
 *   the outline is placed.
 * @property {HTMLDivElement} attachmentsView - The container in which
 *   the attachments are placed.
 * @property {HTMLDivElement} layersView - The container in which
 *   the layers are placed.
 */
export class PDFSidebar {
    /**
     * @param {PDFSidebarOptions} options
     */
    constructor({ elements, pdfViewer, pdfThumbnailViewer, eventBus, l10n, disableNotification, }: PDFSidebarOptions);
    isOpen: boolean;
    active: number;
    isInitialViewSet: boolean;
    /**
     * Callback used when the sidebar has been opened/closed, to ensure that
     * the viewers (PDFViewer/PDFThumbnailViewer) are updated correctly.
     */
    onToggled: any;
    pdfViewer: any;
    pdfThumbnailViewer: any;
    outerContainer: HTMLDivElement;
    viewerContainer: HTMLDivElement;
    toggleButton: HTMLButtonElement;
    thumbnailButton: HTMLButtonElement;
    outlineButton: HTMLButtonElement;
    attachmentsButton: HTMLButtonElement;
    layersButton: HTMLButtonElement;
    thumbnailView: HTMLDivElement;
    outlineView: HTMLDivElement;
    attachmentsView: HTMLDivElement;
    layersView: HTMLDivElement;
    eventBus: any;
    l10n: any;
    _disableNotification: boolean;
    reset(): void;
    /**
     * @type {number} One of the values in {SidebarView}.
     */
    get visibleView(): number;
    get isThumbnailViewVisible(): boolean;
    get isOutlineViewVisible(): boolean;
    get isAttachmentsViewVisible(): boolean;
    get isLayersViewVisible(): boolean;
    /**
     * @param {number} view - The sidebar view that should become visible,
     *                        must be one of the values in {SidebarView}.
     */
    setInitialView(view?: number): void;
    /**
     * @param {number} view - The sidebar view that should be switched to,
     *                        must be one of the values in {SidebarView}.
     * @param {boolean} [forceOpen] - Ensure that the sidebar is open.
     *                                The default value is `false`.
     */
    switchView(view: number, forceOpen?: boolean | undefined): void;
    /**
     * @returns {boolean} Indicating if `this._dispatchEvent` was called.
     * @private
     */
    private _switchView;
    open(): void;
    close(): void;
    toggle(): void;
    /**
     * @private
     */
    private _dispatchEvent;
    /**
     * @private
     */
    private _forceRendering;
    /**
     * @private
     */
    private _updateThumbnailViewer;
    /**
     * @private
     */
    private _showUINotification;
    /**
     * @private
     */
    private _hideUINotification;
    /**
     * @private
     */
    private _addEventListeners;
}
//# sourceMappingURL=pdf_sidebar.d.ts.map